#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <i86.h>
#include <conio.h>

/* Directly accessed gate array registers. */
#define SS_GA_MCTRL     0       /* MIDI control register. */
#define SS_GA_MDATA	1       /* MIDI data register. */
#define SS_GA_HCTRL     2       /* Host control register. */
#define SS_GA_HDATA     3       /* Host data register. */
#define SS_GA_INDEX     4       /* Gate array indirect register index. */
#define SS_GA_DATA      5       /* Gate array indirect register data. */

/* Indirect gate array registers. */
#define SS_GAI_HIS      0       /* Host Interrupt Status Register. */
#define SS_GAI_HIE      1       /* Host Interrupt Enable Register. */
#define SS_GAI_HDAT     2       /* Host DMA A Trigger/Status Register. */
#define SS_GAI_HDBT     3       /* Host DMA B Trigger/Status Register. */
#define SS_GAI_HIC      4       /* Host Interface Interrupt Configuration Register. */
#define SS_GAI_HDC      5       /* Host DMA Interface Configuration Register. */
#define SS_GAI_HCSC     6       /* Host CD-ROM/SCSI Interface Configuration Register. */
#define SS_GAI_HMCA     7       /* Sound Memory Configuration Register A. */
#define SS_GAI_HMCB     8       /* Sound Memory Configuration Register B. */
#define SS_GAI_HMC      9       /* Host Master Control Register. */
#define SS_GAI_HFMA     0xA     /* YM-3812 Emulation Address Register. */
#define SS_GAI_HFMD     0xB     /* YM-3812 Emulation Data Register. */
#define SS_GAI_HFMS     0xC     /* YM-3812 Emulation Status Register. */


typedef struct ss_ctx_s {
    uint32_t    dma_phys_adr;   /* DMA buffer physical address. */
    uint8_t     *dma_buffer;    /* DMA buffer pointer. */
    uint16_t    dma_buf_size;   /* DMA buffer size. */
    uint16_t    io_base_ga;     /* Gate array I/O base (330h default). */
    uint8_t     dma_chan;       /* DMA channel used by adapter. */
    uint8_t     hmc;            /* HMC register saved around FW upload. */
    uint8_t     hcsc;           /* HCSC (codec really) saved around FW upload. */
} ss_ctx;


/* PC DMA registers. */
#define DMAC_BASE_ADDR          0x00    /* DMA base address for channel 0. */
#define DMAC_BASE_COUNT         0x01    /* DMA count for channel 0. */
#define DMAC_16BIT_BASE         0xC0    /* Base of 16-bit DMA controller registers. */

/* Intel 8237A and compatible registers. */
#define DMAC_COMMAND            0x08    /* DMA controller command register. */
#define DMAC_REQ                0x09    /* DMA software request register. */
#define DMAC_MASK_SINGLE        0x0A    /* Single-channel mask register. */
#define DMAC_MODE               0x0B    /* DMA mode register. */
#define DMAC_CLEAR_BPTR         0x0C    /* Clear byte pointer aka flip-flop. */

/* DMA mode register bits. */
#define DMA_MODE_VERIFY         0x00    /* Verify transfer. */
#define DMA_MODE_WRITE          0x04    /* Write transfer. */
#define DMA_MODE_READ           0x08    /* Read transfer. */
#define DMA_MODE_AUTOINIT       0x10    /* Auto-init mode. */
#define DMA_MODE_ADR_INC        0x00    /* Address increment. */
#define DMA_MODE_ADR_DEC        0x20    /* Address decrement. */
#define DMA_MODE_DEMAND         0x00    /* Demand mode. */
#define DMA_MODE_SINGLE         0x40    /* Single mode. */
#define DMA_MODE_BLOCK          0x80    /* Block mode. */
#define DMA_MODE_CASCADE        0xC0    /* Cascade mode. */


/* Mask or unmask DMA for a given channel. */
void dma_mask_set( uint8_t dma_ch, int mask )
{
    uint8_t     reg_adr = DMAC_MASK_SINGLE;
    uint8_t     reg_val;

    if( dma_ch > 3 ) {
        /* 16-bit DMA. */
        reg_adr = DMAC_16BIT_BASE + reg_adr * 2;
        dma_ch  -= 4;
    }
    reg_val = dma_ch | (mask ? 0x04 : 0);
    outp( reg_adr, reg_val );
}


/* Clear the byte pointer aka flip-flop. */
void dma_byteptr_clear( uint8_t dma_ch )
{
    uint8_t     reg_adr = DMAC_CLEAR_BPTR;

    if( dma_ch > 3 )    /* 16-bit DMA. */
        reg_adr = DMAC_16BIT_BASE + reg_adr * 2;

    /* Any write to the register clears the byte pointer. */
    outp( reg_adr, 0 );
}


/* Program the DMA mode for a given channel. */
void dma_mode_set( uint8_t dma_ch, uint8_t mode )
{
    uint8_t     reg_adr = DMAC_MODE;
    uint8_t     reg_val;

    if( dma_ch > 3 ) {
        /* 16-bit DMA. */
        reg_adr = DMAC_16BIT_BASE + reg_adr * 2;
        dma_ch  -= 4;
    }
    reg_val = dma_ch | mode;
    outp( reg_adr, reg_val );
}


/* DMA page register map. The ordering is non-obvious. */
static uint8_t dma_page_reg_map[] = { 0x87, 0x83, 0x81, 0x82 };

/* Set up the base address and count for a transfer. */
void dma_setup_xfer( uint8_t dma_ch, uint32_t phys_addr, uint16_t count )
{
    uint8_t     reg_addr;
    uint8_t     reg_count;
    uint8_t     reg_page;

    if( dma_ch > 3 ) {
        /* 16-bit DMA. */
        reg_addr  = DMAC_BASE_ADDR  + DMAC_16BIT_BASE + (dma_ch & 3) * 4;
        reg_count = DMAC_BASE_COUNT + DMAC_16BIT_BASE + (dma_ch & 3) * 4;
        reg_page  = dma_page_reg_map[dma_ch & 3] + 8;
    } else {
        /* 8-bit DMA. */
        reg_addr  = DMAC_BASE_ADDR  + dma_ch * 2;
        reg_count = DMAC_BASE_COUNT + dma_ch * 2;
        reg_page  = dma_page_reg_map[dma_ch & 3];
    }

    /* Clear the flip-flop. */
    dma_byteptr_clear( dma_ch );

    /* Write the base address. */
    outp( reg_addr, (uint8_t)phys_addr );
    phys_addr >>= 8;
    outp( reg_addr, (uint8_t)phys_addr );
    phys_addr >>= 8;
    outp( reg_page, (uint8_t)phys_addr );

    /* Write the initial count. */
    --count;    /* Need to program actual count - 1. */
    outp( reg_count, (uint8_t)count );
    count >>= 8;
    outp( reg_count, (uint8_t)count );
}


/* Read the current DMA count for a given channel. */
uint16_t dma_read_count( uint8_t dma_ch )
{
    uint8_t     reg_adr;
    uint16_t    dma_count;

    if( dma_ch > 3 ) {
        /* 16-bit DMA. */
        reg_adr = DMAC_BASE_COUNT + DMAC_16BIT_BASE + (dma_ch & 3) * 4;
    } else {
        /* 8-bit DMA. */
        reg_adr = DMAC_BASE_COUNT + dma_ch * 2;
    }

    /* Clear the flip-flop. */
    dma_byteptr_clear( dma_ch );

    /* Read the count (LSB then MSB). */
    dma_count = inp( reg_adr );
    dma_count = dma_count | (inp( reg_adr ) << 8 );

    return( dma_count + 1 );
}


/* Read from an indirect indexed gate array (ODIE/OPUS) register. */
uint8_t ss_ind_rd( ss_ctx *ctx, uint8_t reg )
{
    uint8_t     val;

    outp( ctx->io_base_ga + SS_GA_INDEX, reg );
    val = inp( ctx->io_base_ga + SS_GA_DATA );
    return( val );
}


/* Write to an indirect indexed gate array (ODIE/OPUS) register. */
void ss_ind_wr( ss_ctx *ctx, uint8_t reg, uint8_t val )
{
    outp( ctx->io_base_ga + SS_GA_INDEX, reg );
    outp( ctx->io_base_ga + SS_GA_DATA,  val );
}


/* Read from a direct gate array (ODIE/OPUS) register. */
uint8_t ss_reg_rd( ss_ctx *ctx, uint8_t reg )
{
    uint8_t     val;

    val = inp( ctx->io_base_ga + reg );
    return( val );
}


/* Write to a direct gate array (ODIE/OPUS) register. */
void ss_reg_wr( ss_ctx *ctx, uint8_t reg, uint8_t val )
{
    outp( ctx->io_base_ga + reg, val );
}


/* Detect a Soundscape gate array (ODIE/OPUS). Attempts to preserve
 * the hardware state.
 */
int ss_detect( ss_ctx *ctx )
{
    uint8_t     old_val;

    /* Read the gate array indirect index register. The top 4 bits are
     * not writable and always read back as zeros.
     */
    old_val = ss_reg_rd( ctx, SS_GA_INDEX );
    if( old_val & 0xF0 ) {
        /* Top nibble is set, can't be ODIE etc. */
        return( 0 );
    }

    /* Writing 0xFF must read back as 0x0F. */
    ss_reg_wr( ctx, SS_GA_INDEX, 0xFF );
    if( ss_reg_rd( ctx, SS_GA_INDEX ) != 0x0F ) {
        ss_reg_wr( ctx, SS_GA_INDEX, old_val );
        return( 0 );
    }

    /* Writing 0x00 must read back as 0x00. */
    ss_reg_wr( ctx, SS_GA_INDEX, 0x00 );
    if( ss_reg_rd( ctx, SS_GA_INDEX ) != 0x00 ) {
        ss_reg_wr( ctx, SS_GA_INDEX, old_val );
        return( 0 );
    }

    /* Now check a couple of indirect registers. */
    if( ss_ind_rd( ctx, SS_GAI_HIE ) & 0x0F ) {
        /* The low 4 bits of HIE are always zero. */
        ss_reg_wr( ctx, SS_GA_INDEX, old_val );
        return( 0 );
    }

    if( ss_ind_rd( ctx, SS_GAI_HDC ) & 0x0F ) {
        /* The low 4 bits of HDC are also always zero. */
        ss_reg_wr( ctx, SS_GA_INDEX, old_val );
        return( 0 );
    }

    ss_reg_wr( ctx, SS_GA_INDEX, old_val );

    return( 1 );
}


/* Possible I/O base addresses for classic Soundscape boards. */
static uint16_t ss_io_base[] = { 0x330, 0x340, 0x350, 0x320, 0 };

/* Attempt to find a Soundscape card. */
int ss_find_card( ss_ctx *ctx )
{
    int     idx = 0;
    int     found = 0;

    do {
        ctx->io_base_ga = ss_io_base[idx];
        if( ss_detect( ctx ) ) {
            printf( "Found Soundscape at I/O address %Xh\n", ctx->io_base_ga );
            found = 1;
            break;
        }
    } while( ss_io_base[idx] );

    return( found );
}


/* Program the system DMA controller for transferring a buffer. */
void dma_setup( ss_ctx *ctx, uint32_t phys_addr, uint16_t length, uint8_t mode )
{
    dma_mask_set( ctx->dma_chan, 1 );
    dma_mode_set( ctx->dma_chan, mode );
    dma_setup_xfer( ctx->dma_chan, phys_addr, length );
    dma_mask_set( ctx->dma_chan, 0 );
}


/* Kick of DMA A transfer. */
void ss_dma_a_go( ss_ctx *ctx )
{
    uint8_t     val;

    /* Set up the DMA channel. */
    val = ctx->dma_chan << 4;
    /* Set the 8-bit DMA indicator as appropriate. */
    if( ctx->dma_chan < 4 )
        val |= 0x80;

    //@todo: Make sure bit 0 gets cleared during reset
    /* Set trigger bit to arm DMA transfer. */
    ss_ind_wr( ctx, SS_GAI_HDAT, val | 1);
    /* Clear trigger bit to actually start DMA. */
    ss_ind_wr( ctx, SS_GAI_HDAT, val);
}


/* Prepare the hardware for firmware upload. */
void ss_fw_upload_prep( ss_ctx *ctx )
{
    unsigned    delay_cnt;

    /* Save off the Host Master Control Register. */
    ctx->hmc = ss_ind_rd( ctx, SS_GAI_HMC );

    /* Save the CD-ROM (in fact codec) config register. */
    ctx->hcsc = ss_ind_rd( ctx, SS_GAI_HCSC );

    /* Subsystem reset (bits 6-7 set to zero). */
    ss_ind_wr( ctx, SS_GAI_HMC, 0 );

    /* Delay for a bit. */
    //@todo: Kinda dumb, we should watch some bit?
    for( delay_cnt = 10000; delay_cnt > 0; --delay_cnt )
        ss_ind_rd( ctx, SS_GAI_HMC );

    /* Enable code download mode. The OBP is suspended but DMA enabled. */
    ss_ind_wr( ctx, SS_GAI_HMC, 0x80 );
}


/* Upload a block of firmware. The firmware can be larger than a single
 * DMA transfer (128K firmware, DMA may be 8-bit with 64K limit) and
 * therefore it may be necessary to upload the OBP firmware in multiple
 * blocks. Due to alignment etc. issues, the buffer would typically be
 * 8-32K in size.
 */
int ss_download_blk( ss_ctx *ctx, uint16_t fw_size )
{
    int         success = 0;
    uint16_t    delay_cnt;
    uint16_t    dma_count;
    uint8_t     dma_mode;

    /* Set up the system DMA controller for transfer. */
    dma_mode = DMA_MODE_WRITE | DMA_MODE_SINGLE;
    dma_setup( ctx, ctx->dma_phys_adr, fw_size, dma_mode );
//    printf( "Initial DMA count: %u\n", dma_read_count( ctx->dma_chan ) );
    /* Kick off the DMA on the board. */
    ss_dma_a_go( ctx );

    /* Check for transfer completion, with timeout. */

    delay_cnt = 30000;
    while( delay_cnt-- > 0)
    {
//@todo
//        if (HZ / 50)
//            sleep(HZ / 50);

        dma_count = dma_read_count( ctx->dma_chan );
        if( !dma_count ) {
            success = 1;
            break;
        }
    }

    return( success );
}


/* After FW has been uploaded, start the OBP and verify that the
 * firmware has been successfully initialized.
 */
int ss_fw_upload_done( ss_ctx *ctx )
{
    int         success = 0;
    unsigned    delay_cnt;
    uint8_t     val;

    /* Clear the host communication register. Enables host mode. */
    ss_reg_wr( ctx, SS_GA_HCTRL, 0 );
    ss_reg_wr( ctx, SS_GA_MDATA, 0 );

    /* Enable normal operation and get the OBP going. */
    val = ss_ind_rd( ctx, SS_GAI_HMC );
    val = (val & 0x3F) | 0x40;
    ss_ind_wr( ctx, SS_GAI_HMC, val );

delay( 100 );
    /* Give the OBP a little bit to start running. The OBP firmware will
     * write 0xFF to the host data register as soon as it gets going. Once
     * the firmware is successfully initialized, it writes 0xFE.
     */
    delay_cnt = 10000;
    do {
//@todo        sleep(1);
        val = ss_reg_rd( ctx, SS_GA_HDATA );
        if( val == 0xFF || val == 0xFE ) {
            printf( "OBP response: %02X\n", val );
            success = 1;
            break;
        }
    } while( --delay_cnt > 0 );

    /* Restore the HCSC (codec) state. */
    ss_ind_wr( ctx, SS_GAI_HCSC, ctx->hcsc );
    if( !success ) {
        printf( "Error: No response from OBP after reset!\n" );
        return( 0 );
    }

    success = 0;
    delay_cnt = 10000;
    do {
//@todo        sleep(1);
        if( ss_reg_rd( ctx, SS_GA_HDATA ) == 0xFE ) {   /* Host startup acknowledge */
            success = 1;
            break;
        }
    } while( --delay_cnt > 0 );

    if( success ) {
        printf( "Soundscape OBP initialized\n" );
	printf( "OBP response: %02X\n", ss_reg_rd( ctx, SS_GA_HDATA ) );
    } else
        printf( "Error: OBP initialization failed!\n" );

//        set_control(devc, CTL_MASTER_VOL, 100);
//        set_control(devc, CTL_SYNTH_VOL, 100);

    return( success );
}


uint32_t adr_begin = 0x400000;
uint32_t adr_end   = 0x600000;

int ss_dump_mem( ss_ctx *ctx, const char *fname )
{
    int         success = 0;
    uint32_t    cur_adr = 0;
    uint16_t    length;
    FILE        *f;

    adr_begin = 0;
    adr_end   = 0x20000;

    /* Open the dump file. */
    f = fopen( fname, "wb" );
    if( !f ) {
        printf( "Error: Failed to open dump file `%s'!\n", fname );
        return( success );
    }

    /* Reset the OBP and prepare for FW upload. */
    ss_fw_upload_prep( ctx );

    length = ctx->dma_buf_size;
    /* Now read and DMA over as many blocks as necessary. */
    do {
        /* Download this memory block. */
	printf( "DMA from %08lX\n", cur_adr );
        success = ss_download_blk( ctx, ctx->dma_buf_size );

        if( success && cur_adr >= adr_begin ) {
            /* Read the next block. If the size is zero, we're likely done. */
            length = fwrite( ctx->dma_buffer, 1, ctx->dma_buf_size, f );
            if( length != ctx->dma_buf_size ) {
                success = 0;
                break;
            }
	    printf( "Wrote %u bytes from %08lX\n", length, cur_adr );
        }
        cur_adr += ctx->dma_buf_size;
	if( cur_adr >= adr_end )
	    break;
    } while( length == ctx->dma_buf_size );

    /* Finally start the OBP and check for success report from firmware. */
    if( success )
        success = ss_fw_upload_done( ctx );

    fclose( f );
    return( success );
}

/* ODIE works with IRQ0-3. This maps to board-specific interrupts on the host. */
uint8_t int_map[] = { 2, 7, 12, 15 };


/* Initialize the gate array (ODIE). */
void ss_init_ga( ss_ctx *ctx )
{
#ifndef SSCAPE_REGS
    /*
     * Config register values for Spea/V7 Media FX and Ensoniq S-2000.
     * These values are card
     * dependent. If you have another SoundScape based card, you have to
     * find the correct values. Do the following:
     *  - Compile this driver with SSCAPE_DEBUG1 defined.
     *  - Shut down and power off your machine.
     *  - Boot with DOS so that the SSINIT.EXE program is run.
     *  - Warm boot to {Linux|SYSV|BSD} and write down the lines displayed
     *    when detecting the SoundScape.
     *  - Modify the following list to use the values printed during boot.
     *    Undefine the SSCAPE_DEBUG1
     */
#define SSCAPE_REGS { \
/* I0 */    0x00, \
/* I1 */    0xf0, /* Note! Ignored. Set always to 0xf0 */ \
/* I2 */    0x20, /* Note! Ignored. Set always to 0x20 */ \
/* I3 */    0x20, /* Note! Ignored. Set always to 0x20 */ \
/* I4 */    0xf5, /* Ignored */ \
/* I5 */    0x50, \
/* I6 */    0x00, \
/* I7 */    0x2e, /* I7 MEM config A. Likely to vary between models */ \
/* I8 */    0x00, /* I8 MEM config B. Likely to vary between models */ \
/* I9 */    0x40 /* Ignored */ \
    }
#endif

    static uint8_t  regs[10] = SSCAPE_REGS;
    uint8_t         irq_bits = 4; /* IRQ 5 or so */

    /* Enable all interrupts. */    //@todo we do that below too?!
    ss_ind_wr( ctx, SS_GAI_HIE, 0xF0 );
    /* Disable DMA A. */
    ss_ind_wr( ctx, SS_GAI_HDAT, 0x20 );
    /* Disable DMA B. */
    ss_ind_wr( ctx, SS_GAI_HDBT, 0x20 );    /* DMA channel disabled */
    /* Set up the host interrupt configuration, ISA style. */
    ss_ind_wr( ctx, SS_GAI_HIC, 0xF0 | (irq_bits << 2) | irq_bits );
    /* Don't destroy CD-ROM DMA config bits (0xc0) */
    ss_ind_wr( ctx, SS_GAI_HDC, (regs[5] & 0x3f) | (ss_ind_rd( ctx, SS_GAI_HDC ) & 0xc0 ) );
    /* CD-ROM config (AD1848 codec actually). */
    ss_ind_wr( ctx, SS_GAI_HCSC, regs[6] );
    ss_ind_wr( ctx, SS_GAI_HMCA, regs[7] );
    ss_ind_wr( ctx, SS_GAI_HMCB, regs[8] );
    /* In the HMC, keep CD-ROM and subsystem control, disable SB/FM. */
    ss_ind_wr( ctx, SS_GAI_HMC, (ss_ind_rd( ctx, SS_GAI_HMC ) & 0xf0) | 0x08 );

    /* Enable master interrupt (but not others). */
    ss_ind_wr( ctx, SS_GAI_HIE, 0x80 );
}

/* Physical DMA buffer allocation size. */
#define PHYS_DMA_ALLOC      16384
/* Physical DMA buffer usable size (worst case, really). */
#define PHYS_DMA_BUF_SIZE   (PHYS_DMA_ALLOC / 2)

int main( int argc, char **argv )
{
    ss_ctx      ctx;
    void        *dma_buf;
    uint32_t    dma_phys_adr;

    if( argc < 2 ) {
        printf( "usage: %s <dump_file>\n", argv[0] );
        return( 1 );
    }

    /* Set up a physical DMA buffer, DOS style. Allocate 16K, that guarantees
     * we get 8K buffer not crossing a 64K boundary.
     */
    dma_buf = malloc( PHYS_DMA_ALLOC );
    if( !dma_buf ) {
        printf( "Error: Failed to allocate physical DMA buffer!\n" );
        return( 2 );
    }
    ctx.dma_buffer = dma_buf;

    /* Calculate the physical address of the buffer. */
    dma_phys_adr = ((uint32_t)FP_SEG( dma_buf ) << 4) + FP_OFF( dma_buf );
    printf( "DMA buffer at %04X:%04X, physical %08lX\n",
            FP_SEG( dma_buf ), FP_OFF( dma_buf ), dma_phys_adr );

    /* If the first half of the buffer crosses 64K boundary, use the other half. */
    if( (dma_phys_adr >> 16) != ((dma_phys_adr + PHYS_DMA_BUF_SIZE) >> 16)) {
        dma_phys_adr   += PHYS_DMA_BUF_SIZE;
        ctx.dma_buffer += PHYS_DMA_BUF_SIZE;
        printf( "Aligned DMA buffer at physical %08lX\n", dma_phys_adr );
    }
    ctx.dma_phys_adr = dma_phys_adr;
    ctx.dma_buf_size = PHYS_DMA_BUF_SIZE;

    /* We are going to assume that DMA channel 1 is available. */
    ctx.dma_chan = 1;

    /* Look for a Soundscape board. */
    if( !ss_find_card( &ctx ) ) {
        printf( "Error: No Soundscape found!\n" );
        return( 3 );
    }

    /* Initialize the gate array. */
    ss_init_ga( &ctx );

    /* Dump ourselves some memory. */
    if( !ss_dump_mem( &ctx, argv[1] ) ) {
        printf( "Error: Memory dump failed!\n" );
        return( 4 );
    }

    free( dma_buf );

    return( 0 );
}
