#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <i86.h>
#include <conio.h>

/* Directly accessed gate array registers. */
#define SS_GA_MCTRL     0       /* MIDI control register. */
#define SS_GA_MDATA	1       /* MIDI data register. */
#define SS_GA_HCTRL     2       /* Host control register. */
#define SS_GA_HDATA     3       /* Host data register. */
#define SS_GA_INDEX     4       /* Gate array indirect register index. */
#define SS_GA_DATA      5       /* Gate array indirect register data. */

/* Indirect gate array registers. */
#define SS_GAI_HIS      0       /* Host Interrupt Status Register. */
#define SS_GAI_HIE      1       /* Host Interrupt Enable Register. */
#define SS_GAI_HDAT     2       /* Host DMA A Trigger/Status Register. */
#define SS_GAI_HDBT     3       /* Host DMA B Trigger/Status Register. */
#define SS_GAI_HIC      4       /* Host Interface Interrupt Configuration Register. */
#define SS_GAI_HDC      5       /* Host DMA Interface Configuration Register. */
#define SS_GAI_HCSC     6       /* Host CD-ROM/SCSI Interface Configuration Register. */
#define SS_GAI_HMCA     7       /* Sound Memory Configuration Register A. */
#define SS_GAI_HMCB     8       /* Sound Memory Configuration Register B. */
#define SS_GAI_HMC      9       /* Host Master Control Register. */
#define SS_GAI_HFMA     0xA     /* YM-3812 Emulation Address Register. */
#define SS_GAI_HFMD     0xB     /* YM-3812 Emulation Data Register. */
#define SS_GAI_HFMS     0xC     /* YM-3812 Emulation Status Register. */


typedef struct ss_ctx_s {
    uint32_t    dma_phys_adr;   /* DMA buffer physical address. */
    uint8_t     *dma_buffer;    /* DMA buffer pointer. */
    uint16_t    dma_buf_size;   /* DMA buffer size. */
    uint16_t    io_base_ga;     /* Gate array I/O base (330h default). */
    uint8_t     dma_chan;       /* DMA channel used by adapter. */
    uint8_t     hmc;            /* HMC register saved around FW upload. */
    uint8_t     hcsc;           /* HCSC (codec really) saved around FW upload. */
} ss_ctx;


/* Read from an indirect indexed gate array (ODIE/OPUS) register. */
uint8_t ss_ind_rd( ss_ctx *ctx, uint8_t reg )
{
    uint8_t     val;

    outp( ctx->io_base_ga + SS_GA_INDEX, reg );
    val = inp( ctx->io_base_ga + SS_GA_DATA );
    return( val );
}


/* Write to an indirect indexed gate array (ODIE/OPUS) register. */
void ss_ind_wr( ss_ctx *ctx, uint8_t reg, uint8_t val )
{
    outp( ctx->io_base_ga + SS_GA_INDEX, reg );
    outp( ctx->io_base_ga + SS_GA_DATA,  val );
}


/* Read from a direct gate array (ODIE/OPUS) register. */
uint8_t ss_reg_rd( ss_ctx *ctx, uint8_t reg )
{
    uint8_t     val;

    val = inp( ctx->io_base_ga + reg );
    return( val );
}


/* Write to a direct gate array (ODIE/OPUS) register. */
void ss_reg_wr( ss_ctx *ctx, uint8_t reg, uint8_t val )
{
    outp( ctx->io_base_ga + reg, val );
}


/* Detect a Soundscape gate array (ODIE/OPUS). Attempts to preserve
 * the hardware state.
 */
int ss_detect( ss_ctx *ctx )
{
    uint8_t     old_val;

    /* Read the gate array indirect index register. The top 4 bits are
     * not writable and always read back as zeros.
     */
    old_val = ss_reg_rd( ctx, SS_GA_INDEX );
    if( old_val & 0xF0 ) {
        /* Top nibble is set, can't be ODIE etc. */
        return( 0 );
    }

    /* Writing 0xFF must read back as 0x0F. */
    ss_reg_wr( ctx, SS_GA_INDEX, 0xFF );
    if( ss_reg_rd( ctx, SS_GA_INDEX ) != 0x0F ) {
        ss_reg_wr( ctx, SS_GA_INDEX, old_val );
        return( 0 );
    }

    /* Writing 0x00 must read back as 0x00. */
    ss_reg_wr( ctx, SS_GA_INDEX, 0x00 );
    if( ss_reg_rd( ctx, SS_GA_INDEX ) != 0x00 ) {
        ss_reg_wr( ctx, SS_GA_INDEX, old_val );
        return( 0 );
    }

    /* Now check a couple of indirect registers. */
    if( ss_ind_rd( ctx, SS_GAI_HIE ) & 0x0F ) {
        /* The low 4 bits of HIE are always zero. */
        ss_reg_wr( ctx, SS_GA_INDEX, old_val );
        return( 0 );
    }

    if( ss_ind_rd( ctx, SS_GAI_HDC ) & 0x0F ) {
        /* The low 4 bits of HDC are also always zero. */
        ss_reg_wr( ctx, SS_GA_INDEX, old_val );
        return( 0 );
    }

    ss_reg_wr( ctx, SS_GA_INDEX, old_val );

    return( 1 );
}


/* Possible I/O base addresses for classic Soundscape boards. */
static uint16_t ss_io_base[] = { 0x330, 0x340, 0x350, 0x320, 0 };

/* Attempt to find a Soundscape card. */
int ss_find_card( ss_ctx *ctx )
{
    int     idx = 0;
    int     found = 0;

    do {
        ctx->io_base_ga = ss_io_base[idx];
        if( ss_detect( ctx ) ) {
            printf( "Found Soundscape at I/O address %Xh\n", ctx->io_base_ga );
            found = 1;
            break;
        }
    } while( ss_io_base[idx] );

    return( found );
}


int main( int argc, char **argv )
{
    ss_ctx      ctx;
    uint8_t     reg;
    static char *ind_name[] = { "HIS", "HIE", "HDAT", "HDBT", "HIC", "HDC",
	                        "HCSC", "HMCA", "HMCB", "HMC" };

    /* Look for a Soundscape board. */
    if( !ss_find_card( &ctx ) ) {
        printf( "Error: No Soundscape found!\n" );
        return( 3 );
    }

    for( reg = 0; reg <= SS_GAI_HMC; ++reg ) {
	printf( "%02u %4s: %02X\n", reg, ind_name[reg],
	        ss_ind_rd( &ctx, reg ) );
    }
    return( 0 );
}
